/*
 * Jean-Luc Béchennec pour LOCODUINO
 * 
 *  Gestion d'une gare cachée, les objets, la recherche d'un chemin
 */

/*
 * Identifiant des éléments de voie
 */
enum {
  VOIE_ENTREE,
  VOIE_SORTIE,
  AIG_ENTREE_0,
  AIG_ENTREE_1,
  AIG_ENTREE_2,
  AIG_SORTIE_0,
  AIG_SORTIE_1,
  AIG_SORTIE_2,
  VOIE_GARE_0,
  VOIE_GARE_1,
  VOIE_GARE_2,
  VOIE_GARE_3
};

/*
 * Affichage en clair du non d'un élément de voie, uniquement pour les aiguillages
 */
void afficheVoie(byte identifiant)
{
  switch (identifiant) {
    case AIG_ENTREE_0:
      Serial.print("Aiguillage E0"); break;
    case AIG_ENTREE_1:
      Serial.print("Aiguillage E1"); break;
    case AIG_ENTREE_2:
      Serial.print("Aiguillage E2"); break;
    case AIG_SORTIE_0:
      Serial.print("Aiguillage S0"); break;
    case AIG_SORTIE_1:
      Serial.print("Aiguillage S1"); break;
    case AIG_SORTIE_2:
      Serial.print("Aiguillage S2"); break;
    default:
      Serial.print("** INCONNU **"); break;
  }
}

/*
 * Classe de base des éléments de voie
 * 
 * Fonction virtuelle de recherche de chemin et identifiant
 */
class Voie
{
  private:
    byte mIdentifiant;
 
  public:
    Voie(byte identifiant) { mIdentifiant = identifiant; }
    byte idVoie() { return mIdentifiant; }
    virtual bool creeCheminVers(byte identifiant) = 0;
};

/*
 * Classe pour les voies situées aux extremités
 */
class VoieExtreme : public Voie
{
  private:
    Voie *mVoie;

  public:
    VoieExtreme(byte identifiant) : Voie(identifiant) { mVoie = NULL; }
    void connecte(Voie *voie) { mVoie = voie; }
    virtual bool creeCheminVers(byte identifiant);
};

bool VoieExtreme::creeCheminVers(byte identifiant)
{
  if (identifiant == idVoie()) {
    return true;
  }
  else {
    if (mVoie != NULL) {
      return mVoie->creeCheminVers(identifiant);
    }
    else {
      return false;
    }
  }
}

/*
 * Classe pour les voies de gare
 */
class VoieGare : public Voie
{
  public:
    VoieGare(byte identifiant) : Voie(identifiant) {}
    virtual bool creeCheminVers(byte identifiant);
};

bool VoieGare::creeCheminVers(byte identifiant)
{
  if (identifiant == idVoie()) {
    return true;
  }
  else {
    return false;
  }
}

/*
 * Classe pour les aiguillages
 */
class Aiguillage : public Voie
{
  private:
    Voie *mVoieDroite; /* Voie connectée coté dévié */
    Voie *mVoieDeviee; /* Voie connectée coté droit */
  
  public:
    Aiguillage(byte identifiant) : Voie(identifiant) { mVoieDroite = mVoieDeviee = NULL; }
    void connecteDroite(Voie *voie) { mVoieDroite = voie; }
    void connecteDeviee(Voie *voie) { mVoieDeviee = voie; }
    virtual bool creeCheminVers(byte identifiant);
};

bool Aiguillage::creeCheminVers(byte identifiant)
{
  if (identifiant == idVoie()) {
    return true;
  }
  else {
    if (mVoieDroite != NULL && mVoieDeviee != NULL) {
      /* Uniquement si correctement connecte     */
      /* Cherche la destination cote voie droite */
      if (mVoieDroite->creeCheminVers(identifiant)) {
        /* Destination trouvee cote droit */
        afficheVoie(idVoie());
        Serial.println(": Droit");
        return true;
      }
      else {
        /* Cherche la destination cote voie deviee */
        if (mVoieDeviee->creeCheminVers(identifiant)) {
          /* destination trouvee cote devie */
          afficheVoie(idVoie());
          Serial.println(": Devie");
          return true;
        }
        else {
          /* destination non trouvee */
          return false;
        }
      }
    }
  }
}

/*
 * Instantiation des objets de la gare
 */
VoieExtreme voieEntree(VOIE_ENTREE);
VoieExtreme voieSortie(VOIE_SORTIE);
Aiguillage aiguillageE0(AIG_ENTREE_0);
Aiguillage aiguillageE1(AIG_ENTREE_1);
Aiguillage aiguillageE2(AIG_ENTREE_2);
Aiguillage aiguillageS0(AIG_SORTIE_0);
Aiguillage aiguillageS1(AIG_SORTIE_1);
Aiguillage aiguillageS2(AIG_SORTIE_2);
VoieGare voieGare0(VOIE_GARE_0);
VoieGare voieGare1(VOIE_GARE_1);
VoieGare voieGare2(VOIE_GARE_2);
VoieGare voieGare3(VOIE_GARE_3);

/*
 * Initialisation et test
 */
void setup() {
  /* Connexion des elements de voie du grill d'entree */
  voieEntree.connecte(&aiguillageE0);
  aiguillageE0.connecteDroite(&voieGare0);
  aiguillageE0.connecteDeviee(&aiguillageE1);
  aiguillageE1.connecteDroite(&aiguillageE2);
  aiguillageE1.connecteDeviee(&voieGare1);
  aiguillageE2.connecteDroite(&voieGare3);
  aiguillageE2.connecteDeviee(&voieGare2);
  /* Connexion des elements de voie du grill de sortie */
  voieSortie.connecte(&aiguillageS0);
  aiguillageS0.connecteDroite(&voieGare3);
  aiguillageS0.connecteDeviee(&aiguillageS1);
  aiguillageS1.connecteDroite(&aiguillageS2);
  aiguillageS1.connecteDeviee(&voieGare2);
  aiguillageS2.connecteDroite(&voieGare0);
  aiguillageS2.connecteDeviee(&voieGare1);

  Serial.begin(9600);
  Serial.println("***** TEST");
  /* Test des chemins */
  Serial.println("===== Test de l'entree");
  Serial.println("Voie gare 0");
  voieEntree.creeCheminVers(VOIE_GARE_0);
  Serial.println("Voie gare 1");
  voieEntree.creeCheminVers(VOIE_GARE_1);
  Serial.println("Voie gare 2");
  voieEntree.creeCheminVers(VOIE_GARE_2);
  Serial.println("Voie gare 3");
  voieEntree.creeCheminVers(VOIE_GARE_3);
  
  Serial.println("===== Test de la sortie");
  Serial.println("Voie gare 0");
  voieSortie.creeCheminVers(VOIE_GARE_0);
  Serial.println("Voie gare 1");
  voieSortie.creeCheminVers(VOIE_GARE_1);
  Serial.println("Voie gare 2");
  voieSortie.creeCheminVers(VOIE_GARE_2);
  Serial.println("Voie gare 3");
  voieSortie.creeCheminVers(VOIE_GARE_3);
}

void loop() {
  /* Rien pour l'instant */
}
